#!/usr/bin/env python3

"""Module containing the Cpptraj Rms class and the command line interface."""
import argparse
from typing import Optional
from pathlib import PurePath
from biobb_common.generic.biobb_object import BiobbObject
from biobb_common.configuration import settings
from biobb_common.tools import file_utils as fu
from biobb_common.tools.file_utils import launchlogger
from biobb_analysis.ambertools.common import get_default_value, check_top_path, check_traj_path, check_out_path, get_binary_path, get_in_parameters, get_negative_mask, copy_instructions_file_to_container, setup_structure, get_mask, get_reference_rms


class CpptrajRms(BiobbObject):
    """
    | biobb_analysis CpptrajRms
    | Wrapper of the Ambertools Cpptraj module for calculating the Root Mean Square deviation (RMSd) of a given cpptraj compatible trajectory.
    | Cpptraj (the successor to ptraj) is the main program in Ambertools for processing coordinate trajectories and data files. The parameter names and defaults are the same as the ones in the official `Cpptraj manual <https://amber-md.github.io/cpptraj/CPPTRAJ.xhtml>`_.

    Args:
        input_top_path (str): Path to the input structure or topology file. File type: input. `Sample file <https://github.com/bioexcel/biobb_analysis/raw/master/biobb_analysis/test/data/ambertools/cpptraj.parm.top>`_. Accepted formats: top (edam:format_3881), pdb (edam:format_1476), prmtop (edam:format_3881), parmtop (edam:format_3881), zip (edam:format_3987).
        input_traj_path (str): Path to the input trajectory to be processed.  File type: input. `Sample file <https://github.com/bioexcel/biobb_analysis/raw/master/biobb_analysis/test/data/ambertools/cpptraj.traj.dcd>`_. Accepted formats: mdcrd (edam:format_3878), crd (edam:format_3878), cdf (edam:format_3650), netcdf (edam:format_3650), nc (edam:format_3650), restart (edam:format_3886), ncrestart (edam:format_3886), restartnc (edam:format_3886), dcd (edam:format_3878), charmm (edam:format_3887), cor (edam:format_2033), pdb (edam:format_1476), mol2 (edam:format_3816), trr (edam:format_3910), gro (edam:format_2033), binpos (edam:format_3885), xtc (edam:format_3875), cif (edam:format_1477), arc (edam:format_2333), sqm (edam:format_2033), sdf (edam:format_3814), conflib (edam:format_2033).
        input_exp_path (str) (Optional): Path to the experimental reference file (required if reference = experimental). File type: input. `Sample file <https://github.com/bioexcel/biobb_analysis/raw/master/biobb_analysis/test/data/ambertools/experimental.1e5t.pdb>`_. Accepted formats: pdb (edam:format_1476).
        output_cpptraj_path (str): Path to the output processed analysis. File type: output. `Sample file <https://github.com/bioexcel/biobb_analysis/raw/master/biobb_analysis/test/reference/ambertools/ref_cpptraj.rms.first.dat>`_. Accepted formats: dat (edam:format_1637), agr (edam:format_2033), xmgr (edam:format_2033), gnu (edam:format_2033).
        output_traj_path (str) (Optional): Path to the output processed trajectory. File type: output. `Sample file <https://github.com/bioexcel/biobb_analysis/raw/master/biobb_analysis/test/data/ambertools/cpptraj.traj.dcd>`_. Accepted formats: mdcrd (edam:format_3878), crd (edam:format_3878), cdf (edam:format_3650), netcdf (edam:format_3650), nc (edam:format_3650), restart (edam:format_3886), ncrestart (edam:format_3886), restartnc (edam:format_3886), dcd (edam:format_3878), charmm (edam:format_3887), cor (edam:format_2033), pdb (edam:format_1476), mol2 (edam:format_3816), trr (edam:format_3910), gro (edam:format_2033), binpos (edam:format_3885), xtc (edam:format_3875), cif (edam:format_1477), arc (edam:format_2333), sqm (edam:format_2033), sdf (edam:format_3814), conflib (edam:format_2033).
        properties (dic - Python dictionary object containing the tool parameters, not input/output files):
            * **start** (*int*) - (1) [1~100000|1] Starting frame for slicing
            * **end** (*int*) - (-1) [-1~100000|1] Ending frame for slicing
            * **steps** (*int*) - (1) [1~100000|1] Step for slicing
            * **mask** (*str*) - ("all-atoms") Mask definition. Values: c-alpha (All c-alpha atoms; protein only), backbone (Backbone atoms), all-atoms (All system atoms), heavy-atoms (System heavy atoms; not hydrogen), side-chain (All not backbone atoms), solute (All system atoms except solvent atoms), ions (All ion molecules), solvent (All solvent atoms), AnyAmberFromatMask (Amber atom selection syntax like `@*`).
            * **reference** (*str*) - ("first") Reference definition. Values: first (Use the first trajectory frame as reference), average (Use the average of all trajectory frames as reference), experimental (Use the experimental structure as reference).
            * **nofit** (*bool*) - (False) Do not perform best-fit RMSD
            * **norotate** (*bool*) - (False) Translate but do not rotate coordinates
            * **nomod** (*bool*) - (False) Do not modify coordinates
            * **binary_path** (*str*) - ("cpptraj") Path to the cpptraj executable binary.
            * **remove_tmp** (*bool*) - (True) [WF property] Remove temporal files.
            * **restart** (*bool*) - (False) [WF property] Do not execute if output files exist.
            * **sandbox_path** (*str*) - ("./") [WF property] Parent path to the sandbox directory.
            * **container_path** (*str*) - (None) Container path definition.
            * **container_image** (*str*) - ('afandiadib/ambertools:serial') Container image definition.
            * **container_volume_path** (*str*) - ('/tmp') Container volume path definition.
            * **container_working_dir** (*str*) - (None) Container working directory definition.
            * **container_user_id** (*str*) - (None) Container user_id definition.
            * **container_shell_path** (*str*) - ('/bin/bash') Path to default shell inside the container.

    Examples:
        This is a use example of how to use the building block from Python::

            from biobb_analysis.ambertools.cpptraj_rms import cpptraj_rms
            prop = {
                'start': 1,
                'end': -1,
                'steps': 1,
                'mask': 'c-alpha',
                'reference': 'first'
            }
            cpptraj_rms(input_top_path='/path/to/myTopology.top',
                        input_traj_path='/path/to/myTrajectory.dcd',
                        output_cpptraj_path='/path/to/newAnalysis.dat',
                        input_exp_path= '/path/to/myExpStructure.pdb',
                        output_traj_path='/path/to/newTrajectory.netcdf',
                        properties=prop)

    Info:
        * wrapped_software:
            * name: Ambertools Cpptraj
            * version: >=20.0
            * license: GNU
        * ontology:
            * name: EDAM
            * schema: http://edamontology.org/EDAM.owl

    """

    def __init__(self, input_top_path, input_traj_path, output_cpptraj_path,
                 input_exp_path=None, output_traj_path=None, properties=None, **kwargs) -> None:
        properties = properties or {}

        # Call parent class constructor
        super().__init__(properties)
        self.locals_var_dict = locals().copy()

        # Input/Output files
        self.io_dict = {
            "in": {"input_top_path": input_top_path, "input_traj_path": input_traj_path, "input_exp_path": input_exp_path},
            "out": {"output_cpptraj_path": output_cpptraj_path, "output_traj_path": output_traj_path}
        }

        # Properties specific for BB
        self.instructions_file = get_default_value('instructions_file')
        self.start = properties.get('start', 1)
        self.end = properties.get('end', -1)
        self.steps = properties.get('steps', 1)
        self.mask = properties.get('mask', 'all-atoms')
        self.reference = properties.get('reference', 'first')
        self.nofit = properties.get('nofit', False)
        self.norotate = properties.get('norotate', False)
        self.nomod = properties.get('nomod', False)
        self.properties = properties
        self.binary_path = get_binary_path(properties, 'binary_path')

        # Check the properties
        self.check_properties(properties)
        self.check_arguments()

    def check_data_params(self, out_log, err_log):
        """ Checks all the input/output paths and parameters """
        self.io_dict["in"]["input_top_path"], self.input_top_path_orig = check_top_path(self.io_dict["in"]["input_top_path"], out_log, self.__class__.__name__)
        self.io_dict["in"]["input_traj_path"] = check_traj_path(self.io_dict["in"]["input_traj_path"], out_log, self.__class__.__name__)
        self.io_dict["out"]["output_cpptraj_path"] = check_out_path(self.io_dict["out"]["output_cpptraj_path"], out_log, self.__class__.__name__)
        if self.io_dict["out"]["output_traj_path"]:
            self.io_dict["out"]["output_traj_path"] = check_out_path(self.io_dict["out"]["output_traj_path"], out_log, self.__class__.__name__)
        self.in_parameters = {'start': self.start, 'end': self.end, 'step': self.steps, 'mask': self.mask, 'reference': self.reference}

    def create_instructions_file(self, container_io_dict, out_log, err_log):
        """Creates an input file using the properties file settings"""
        instructions_list = []
        # different path if container execution or not
        if self.container_path:
            self.instructions_file = str(PurePath(self.container_volume_path).joinpath(self.instructions_file))
        else:
            self.instructions_file = str(PurePath(fu.create_unique_dir()).joinpath(self.instructions_file))
        fu.create_name(prefix=self.prefix, step=self.step, name=self.instructions_file)

        # parm
        instructions_list.append('parm ' + container_io_dict["in"]["input_top_path"])

        # trajin
        in_params = get_in_parameters(self.in_parameters, out_log)
        instructions_list.append('trajin ' + container_io_dict["in"]["input_traj_path"] + ' ' + in_params)

        # Set up
        instructions_list += setup_structure(self)

        # mask
        mask = self.in_parameters.get('mask', '')
        ref_mask = ''
        if mask:
            strip_mask = get_negative_mask(mask, out_log)
            ref_mask = get_mask(mask, out_log)
            instructions_list.append('strip ' + strip_mask)

        # reference
        reference = self.in_parameters.get('reference', '')
        inp_exp_pth = None
        if "input_exp_path" in container_io_dict["in"]:
            inp_exp_pth = container_io_dict["in"]["input_exp_path"]
        instructions_list += get_reference_rms(reference, container_io_dict["out"]["output_cpptraj_path"], inp_exp_pth, ref_mask, True,
                                               self.__class__.__name__, out_log, self.nofit, self.norotate, self.nomod)

        # trajout
        if ("output_traj_path" in container_io_dict["out"]):
            instructions_list.append('trajout ' + container_io_dict["out"]["output_traj_path"])

        # create .in file
        with open(self.instructions_file, 'w') as mdp:
            for line in instructions_list:
                mdp.write(line.strip() + '\n')

        return self.instructions_file

    @launchlogger
    def launch(self) -> int:
        """Execute the :class:`CpptrajRms <ambertools.cpptraj_rms.CpptrajRms>` ambertools.cpptraj_rms.CpptrajRms object."""

        # check input/output paths and parameters
        self.check_data_params(self.out_log, self.err_log)

        # Setup Biobb
        if self.check_restart():
            return 0
        self.stage_files()

        # create instructions file
        self.create_instructions_file(self.stage_io_dict, self.out_log, self.err_log)

        # if container execution, copy intructions file to container
        if self.container_path:
            copy_instructions_file_to_container(self.instructions_file, self.stage_io_dict['unique_dir'])

        # create cmd and launch execution
        self.cmd = [self.binary_path, '-i', self.instructions_file]

        # Run Biobb block
        self.run_biobb()

        # Copy files to host
        self.copy_to_host()

        # remove temporary folder(s)
        self.tmp_files.extend([
            self.stage_io_dict.get("unique_dir", ""),
            str(PurePath(self.instructions_file).parent)
        ])
        self.remove_tmp_files()

        self.check_arguments(output_files_created=True, raise_exception=False)

        return self.return_code


def cpptraj_rms(input_top_path: str, input_traj_path: str, output_cpptraj_path: str, input_exp_path: Optional[str] = None, output_traj_path: Optional[str] = None, properties: Optional[dict] = None, **kwargs) -> int:
    """Execute the :class:`CpptrajRms <ambertools.cpptraj_rms.CpptrajRms>` class and
    execute the :meth:`launch() <ambertools.cpptraj_rms.CpptrajRms.launch>` method."""

    return CpptrajRms(input_top_path=input_top_path,
                      input_traj_path=input_traj_path,
                      output_cpptraj_path=output_cpptraj_path,
                      input_exp_path=input_exp_path,
                      output_traj_path=output_traj_path,
                      properties=properties, **kwargs).launch()


def main():
    """Command line execution of this building block. Please check the command line documentation."""
    parser = argparse.ArgumentParser(description="Calculates the Root Mean Square deviation (RMSd) of a given cpptraj compatible trajectory.", formatter_class=lambda prog: argparse.RawTextHelpFormatter(prog, width=99999))
    parser.add_argument('--config', required=False, help='Configuration file')

    # Specific args of each building block
    required_args = parser.add_argument_group('required arguments')
    required_args.add_argument('--input_top_path', required=True, help='Path to the input structure or topology file. Accepted formats: top, pdb, prmtop, parmtop, zip.')
    required_args.add_argument('--input_traj_path', required=True, help='Path to the input trajectory to be processed. Accepted formats: crd, cdf, netcdf, restart, ncrestart, restartnc, dcd, charmm, cor, pdb, mol2, trr, gro, binpos, xtc, cif, arc, sqm, sdf, conflib.')
    parser.add_argument('--input_exp_path', required=False, help='Path to the experimental reference file (required if reference = experimental).')
    required_args.add_argument('--output_cpptraj_path', required=True, help='Path to the output processed analysis.')
    parser.add_argument('--output_traj_path', required=False, help='Path to the output processed trajectory.')

    args = parser.parse_args()
    args.config = args.config or "{}"
    properties = settings.ConfReader(config=args.config).get_prop_dic()

    # Specific call of each building block
    cpptraj_rms(input_top_path=args.input_top_path,
                input_traj_path=args.input_traj_path,
                output_cpptraj_path=args.output_cpptraj_path,
                input_exp_path=args.input_exp_path,
                output_traj_path=args.output_traj_path,
                properties=properties)


if __name__ == '__main__':
    main()



from . import cpptraj_average
from . import cpptraj_bfactor
from . import cpptraj_convert
from . import cpptraj_dry
from . import cpptraj_image
from . import cpptraj_mask
from . import cpptraj_rgyr
from . import cpptraj_rms
from . import cpptraj_rmsf
from . import cpptraj_slice
from . import cpptraj_snapshot
from . import cpptraj_strip

name = "ambertools"
__all__ = ["cpptraj_average", "cpptraj_bfactor", "cpptraj_convert", "cpptraj_dry", "cpptraj_image", "cpptraj_mask", "cpptraj_rgyr", "cpptraj_rms", "cpptraj_rmsf", "cpptraj_slice", "cpptraj_snapshot", "cpptraj_strip"]



""" Common functions for package biobb_analysis.ambertools """
from pathlib import Path, PurePath
import zipfile
import shutil
from biobb_common.tools import file_utils as fu


def check_top_path(path, out_log, classname):
    """ Checks topology input file """
    orig_path = path
    if not Path(path).exists():
        fu.log(classname + ': Unexisting topology input file, exiting', out_log)
        raise SystemExit(classname + ': Unexisting topology input file')
    file_extension = PurePath(path).suffix
    if not is_valid_topology(file_extension[1:]):
        fu.log(classname + ': Format %s in topology input file is not compatible' % file_extension[1:], out_log)
        raise SystemExit(classname + ': Format %s in topology input file is not compatible' % file_extension[1:])
    if zipfile.is_zipfile(path):
        top_file = fu.unzip_top(zip_file=path, out_log=out_log)
        path = top_file
    return path, orig_path


def check_traj_path(path, out_log, classname):
    """ Checks trajectory input file """
    if not Path(path).exists():
        fu.log(classname + ': Unexisting trajectory input file, exiting', out_log)
        raise SystemExit(classname + ': Unexisting trajectory input file')
    file_extension = PurePath(path).suffix
    if not is_valid_trajectory(file_extension[1:]):
        fu.log(classname + ': Format %s in trajectory input file is not compatible' % file_extension[1:], out_log)
        raise SystemExit(classname + ': Format %s in trajectory input file is not compatible' % file_extension[1:])
    return path


def check_out_path(path, out_log, classname):
    """ Checks if output folder exists """
    if PurePath(path).parent and not Path(PurePath(path).parent).exists():
        fu.log(classname + ': Unexisting output folder, exiting', out_log)
        raise SystemExit(classname + ': Unexisting output folder')
    return path


def get_parameters(properties, type, classname, out_log):
    """ Gets in_parameters and out_parameters """
    if not properties.get(type, dict()):
        fu.log('No %s parameters provided' % type, out_log)
        return get_default_value(classname)[type]

    return {k: v for k, v in properties.get(type, dict()).items()}


def get_binary_path(properties, type):
    """ Gets binary path """
    return properties.get(type, get_default_value(type))


def check_in_path(path, out_log, classname):
    """ Checks input instructions file """
    if not Path(path).exists():
        fu.log(classname + ': Unexisting input instructions file, exiting', out_log)
        raise SystemExit(classname + ': Unexisting input instructions file')

    # check syntax for instructions file
    syntax_instructions = True
    err_instructions = ''
    if 'parm ' not in open(path).read():
        syntax_instructions = False
        err_instructions += 'No topology provided'
    if 'trajin ' not in open(path).read():
        syntax_instructions = False
        err_instructions += ' No input trajectory provided'
    if not syntax_instructions:
        fu.log(classname + ': Incorrect syntax for instructions file: %s, exiting' % err_instructions, out_log)
        raise SystemExit(classname + ': Incorrect syntax for instructions file: %s, exiting' % err_instructions)


def get_default_value(key):
    """ Gives default values according to the given key """
    default_values = {
        "start": 1,
        "end": -1,
        "step": 1,
        "snapshot": 1,
        "format": "netcdf",
        "mask": "all-atoms",
        "reference": "first",
        "average": "MyAvg",
        "instructions_file": "instructions.in",
        "binary_path": "cpptraj",
        # default conf for Average
        "Average": {
            "in_parameters": {
                "start": 1,
                "end": -1,
                "step": 1,
                "mask": "all-atoms"
            },
            "out_parameters": {
                "format": "pdb"
            }
        },
        # default conf for Bfactor
        "Bfactor": {
            "in_parameters": {
                "start": 1,
                "end": -1,
                "step": 1,
                "mask": "all-atoms",
                "reference": "first"
            }
        },
        # default conf for Convert
        "Convert": {
            "in_parameters": {
                "start": 1,
                "end": -1,
                "step": 1,
                "mask": "all-atoms"
            },
            "out_parameters": {
                "format": "netcdf"
            }
        },
        # default conf for Dry
        "Dry": {
            "in_parameters": {
                "start": 1,
                "end": -1,
                "step": 1,
                "mask": "all-atoms"
            },
            "out_parameters": {
                "format": "netcdf"
            }
        },
        # default conf for Image
        "Image": {
            "in_parameters": {
                "start": 1,
                "end": -1,
                "step": 1,
                "mask": "all-atoms"
            },
            "out_parameters": {
                "format": "netcdf"
            }
        },
        # default conf for Mask
        "Mask": {
            "in_parameters": {
                "start": 1,
                "end": -1,
                "step": 1,
                "mask": "all-atoms"
            },
            "out_parameters": {
                "format": "netcdf"
            }
        },
        # default conf for Rgyr
        "Rgyr": {
            "in_parameters": {
                "start": 1,
                "end": -1,
                "step": 1,
                "mask": "all-atoms"
            }
        },
        # default conf for Rms
        "Rms": {
            "in_parameters": {
                "start": 1,
                "end": -1,
                "step": 1,
                "mask": "all-atoms",
                "reference": "first"
            }
        },
        # default conf for Rmsf
        "Rmsf": {
            "in_parameters": {
                "start": 1,
                "end": -1,
                "step": 1,
                "mask": "all-atoms",
                "reference": "first"
            }
        },
        # default conf for Slice
        "Slice": {
            "in_parameters": {
                "start": 1,
                "end": -1,
                "step": 1,
                "mask": "all-atoms"
            },
            "out_parameters": {
                "format": "netcdf"
            }
        },
        # default conf for Snapshot
        "Snapshot": {
            "in_parameters": {
                "snapshot": 12,
                "mask": "all-atoms"
            },
            "out_parameters": {
                "format": "pdb"
            }
        },
        # default conf for Strip
        "Strip": {
            "in_parameters": {
                "start": 1,
                "end": -1,
                "step": 1,
                "mask": "all-atoms"
            },
            "out_parameters": {
                "format": "netcdf"
            }
        }
    }

    return default_values[key]


def is_valid_topology(ext):
    """ Checks if trajectory format is compatible with Cpptraj """
    formats = 'top', 'pdb', 'prmtop', 'parmtop', 'zip'
    return ext in formats


def is_valid_trajectory(traj):
    """ Checks if trajectory format is compatible with Cpptraj """
    formats = 'mdcrd', 'crd', 'cdf', 'netcdf', 'nc', 'restart', 'ncrestart', 'restartnc', 'dcd', 'charmm', 'cor', 'pdb', 'mol2', 'trr', 'gro', 'binpos', 'xtc', 'cif', 'arc', 'sqm', 'sdf', 'conflib'
    return traj in formats


def is_valid_reference(ref):
    """ Checks if reference is correct """
    references = 'first', 'average', 'experimental'
    return ref in references


def get_mask_atoms(key):
    """ Gives mask atoms according to the given key """
    masks = {
        "c-alpha": "@CA",
        "backbone": "@C,CA,N,O,C3',O3',C4',C5',O5',P",
        "all-atoms": ":*",
        "heavy-atoms": "!@H*,1H*,2H*,3H*",
        "side-chain": "!@CA,C,N,O,H,HA,C3',O3',C4',C5',O5',P",
        "solute": "!:WAT,HOH,SOL,TIP3,TP3,SOD,CLA,Na+,Cl-,NA,CL,K+,K",
        "ions": ":SOD,CLA,Na+,Cl-,NA,CL,K+,K",
        "solvent": ":WAT,HOH,SOL,TIP3,TP3"
    }

    # if key incorrect, return default value and message
    if key in masks:
        return masks[key], None
    else:
        return key, None  # Allow for Amber mask


def get_in_parameters(list, out_log, type='None'):
    """ Return string with input parameters """
    # if strip or mask, no mandatory trajin parameters
    if type == 'strip' or type == 'mask':
        start = '' if 'start' not in list else str(list['start'])
        end = '' if 'end' not in list else str(list['end'])
        step = '' if 'step' not in list else str(list['step'])
        if (not start or start == 'None') and (not end or end == 'None') and (not step or step == 'None'):
            return ''
        else:
            if not start:
                start = str(get_default_value("start"))
                fu.log('No start value provided in configuration file or incorrect format, assigned default value: %s' % get_default_value('start'), out_log)
            if not end:
                end = str(get_default_value("end"))
                fu.log('No end value provided in configuration file or incorrect format, assigned default value: %s' % get_default_value('end'), out_log)
            if not step:
                step = str(get_default_value("step"))
                fu.log('No step value provided in configuration file or incorrect format, assigned default value: %s' % get_default_value('step'), out_log)
    else:
        # check if trajin parameters are provided and have correct format
        if (type == 'snapshot'):
            snapshot = str(get_default_value("snapshot")) if 'snapshot' not in list else str(list['snapshot'])
            if 'snapshot' not in list or not list['snapshot'] or not isinstance(list['snapshot'], int):
                snapshot = str(get_default_value("snapshot"))
                fu.log('No snapshot value provided in configuration file or incorrect format, assigned default value: %s' % get_default_value('snapshot'), out_log)
            start = snapshot
            end = snapshot
            step = '1'
        else:
            start = str(get_default_value("start")) if 'start' not in list else str(list['start'])
            if 'start' not in list or not list['start'] or not isinstance(list['start'], int):
                start = str(get_default_value("start"))
                fu.log('No start value provided in configuration file or incorrect format, assigned default value: %s' % get_default_value('start'), out_log)
            end = str(get_default_value("end")) if 'end' not in list else str(list['end'])
            if 'end' not in list or not list['end'] or not isinstance(list['end'], int):
                end = str(get_default_value("end"))
                fu.log('No end value provided in configuration file or incorrect format, assigned default value: %s' % get_default_value('end'), out_log)
            step = str(get_default_value("step")) if 'step' not in list else str(list['step'])
            if 'step' not in list or not list['step'] or not isinstance(list['step'], int):
                step = str(get_default_value("step"))
                fu.log('No step value provided in configuration file or incorrect format, assigned default value: %s' % get_default_value('step'), out_log)

            # checking start <= end
            if end != '-1' and start > end:
                fu.log('End must be -1 (indicating the end of the trajectory) or more or equal than start. Your values are start: %s, end: %s' % (start, end), out_log)
                raise SystemExit('End must be -1 (indicating the end of the trajectory) or greater or equal than start. Your values are start: %s, end: %s' % (start, end))

    return start + " " + end + " " + step


def setup_structure(out_log):
    """ Sets up the structure """
    instructions_list = []
    mask_atoms = get_mask('heavy-atoms', out_log)
    instructions_list.append('center ' + mask_atoms + ' origin')
    instructions_list.append('autoimage')
    instructions_list.append('rms first ' + mask_atoms)
    mask_solvent = get_mask('solvent', out_log)
    mask_ions = get_mask('ions', out_log)
    instructions_list.append('strip ' + mask_solvent + ',' + mask_ions[1:])

    return instructions_list


def get_negative_mask(key, out_log):
    """ Gives the negative mask according to the given key """
    atoms, msg = get_mask_atoms(key)
    if atoms[0] == '!':
        mask = atoms[1:]
    else:
        mask = '!' + atoms
    # if mask incorrect, give message
    if msg:
        fu.log(msg, out_log)

    return mask


def get_mask(key, out_log):
    """ Gives mask according to the given key """
    mask, msg = get_mask_atoms(key)
    # if mask incorrect, give message
    if msg:
        fu.log(msg, out_log)

    return mask


def get_reference(ref, output_cpptraj_path, input_exp_path, mask, output, classname, out_log):
    """ Gives reference instructions according to the given key """
    instructions_list = []
    if not ref or ref == 'None':
        ref = get_default_value('reference')
        fu.log('No reference provided in configuration file, assigned default value: %s' % get_default_value('reference'), out_log)

    if not is_valid_reference(ref):
        fu.log('Reference %s is not compatible, assigned default value: %s' % (ref, get_default_value('reference')), out_log)
        ref = get_default_value('reference')

    if ref == 'first':
        if output:
            instructions_list.append('rms first out ' + output_cpptraj_path)
        else:
            instructions_list.append('rms first')

    if ref == 'average':
        instructions_list.append('average crdset ' + get_default_value('average'))
        instructions_list.append('run')
        if output:
            instructions_list.append('rms ref ' + get_default_value('average') + ' ' + mask + ' out ' + output_cpptraj_path)
        else:
            instructions_list.append('rms ref ' + get_default_value('average') + ' ' + mask)

    if ref == 'experimental':
        if not input_exp_path:
            fu.log('No experimental structure provided, exiting', out_log)
            raise SystemExit(classname + ': input_exp_path is mandatory')
        instructions_list.append('parm ' + input_exp_path + ' noconect [exp]')
        solute, msg = get_mask_atoms('solute')
        instructions_list.append('reference ' + input_exp_path + ' ' + solute + ' parm [exp]')
        backbone, msg = get_mask_atoms('backbone')
        if output:
            instructions_list.append('rms reference ' + mask + ' out ' + output_cpptraj_path)
        else:
            instructions_list.append('rms reference ' + mask)

    return instructions_list


def get_reference_rms(ref, output_cpptraj_path, input_exp_path, mask, output, classname, out_log, nofit=False, norotate=False, nomod=False):
    """ Gives reference instructions according to the given key """
    instructions_list = []
    if not ref or ref == 'None':
        ref = get_default_value('reference')
        fu.log('No reference provided in configuration file, assigned default value: %s' % get_default_value('reference'), out_log)

    if not is_valid_reference(ref):
        fu.log('Reference %s is not compatible, assigned default value: %s' % (ref, get_default_value('reference')), out_log)
        ref = get_default_value('reference')

    flags = []
    if nofit:
        flags.append("nofit")
    if norotate:
        flags.append("norotate")
    if nomod:
        flags.append("nomod")

    flags_str = " ".join(flags)

    if ref == 'first':
        if output:
            instructions_list.append('rms first out ' + output_cpptraj_path + f' {flags_str}')
        else:
            instructions_list.append('rms first' + f' {flags_str}')

    if ref == 'average':
        instructions_list.append('average crdset ' + get_default_value('average'))
        instructions_list.append('run')
        if output:
            instructions_list.append('rms ref ' + get_default_value('average') + ' ' + mask + ' out ' + output_cpptraj_path + f' {flags_str}')
        else:
            instructions_list.append('rms ref ' + get_default_value('average') + ' ' + mask + f' {flags_str}')

    if ref == 'experimental':
        if not input_exp_path:
            fu.log('No experimental structure provided, exiting', out_log)
            raise SystemExit(classname + ': input_exp_path is mandatory')
        instructions_list.append('parm ' + input_exp_path + ' noconect [exp]')
        solute, msg = get_mask_atoms('solute')
        instructions_list.append('reference ' + input_exp_path + ' ' + solute + ' parm [exp]')
        backbone, msg = get_mask_atoms('backbone')
        if output:
            instructions_list.append('rms reference ' + mask + ' out ' + output_cpptraj_path + f' {flags_str}')
        else:
            instructions_list.append('rms reference ' + mask + f' {flags_str}')

    return instructions_list


def get_out_parameters(list, out_log):
    """ Return string with output parameters """
    format = list['format']
    # msg = None
    # check if format provided
    if not format:
        format = get_default_value('format')
        fu.log('No format provided in configuration file, assigned default value: %s' % get_default_value('format'), out_log)
    # check if valid format
    if not is_valid_trajectory(format):
        fu.log('Format %s is not compatible, assigned default value: %s' % (format, get_default_value('format')), out_log)
        format = get_default_value('format')

    return format


def copy_instructions_file_to_container(instructions_file, unique_dir):
    shutil.copy2(instructions_file, unique_dir)


def remove_tmp_files(list, remove_tmp, out_log, input_top_path_orig=None, input_top_path=None):
    """ Removes temporal files generated by the wrapper """
    tmp_files = list
    if zipfile.is_zipfile(str(input_top_path_orig)):
        tmp_files.append(PurePath(str(input_top_path)).parent)

    if remove_tmp:
        removed_files = [f for f in tmp_files if fu.rm(f)]
        fu.log('Removed: %s' % str(removed_files), out_log)



#!/usr/bin/env python3

# AFEGIR PARÀMETRES DE GMX_IMAGE: PBC, CENTER, UR, FIT (NOMÉS APPEND SI L'USUARI ELS AFEGEIX)

"""Module containing the GMX TrjConvStr class and the command line interface."""
import argparse
from typing import Optional
from biobb_common.generic.biobb_object import BiobbObject
from biobb_common.configuration import settings
from biobb_common.tools import file_utils as fu
from biobb_common.tools.file_utils import launchlogger
from biobb_analysis.gromacs.common import get_binary_path, check_input_path, check_traj_path, check_index_path, get_selection_index_file, get_selection, check_out_traj_path


class GMXTrjConvStr(BiobbObject):
    """
    | biobb_analysis GMXTrjConvStr
    | Wrapper of the GROMACS trjconv module for converting between GROMACS compatible structure file formats and/or extracting a selection of atoms.
    | GROMACS trjconv module can convert trajectory files in many ways. See the `GROMACS trjconv <http://manual.gromacs.org/documentation/2018/onlinehelp/gmx-trjconv.html>`_ official documentation for further information.

    Args:
        input_structure_path (str): Path to the input structure file. File type: input. `Sample file <https://github.com/bioexcel/biobb_analysis/raw/master/biobb_analysis/test/data/gromacs/trajectory.trr>`_. Accepted formats: xtc (edam:format_3875), trr (edam:format_3910), cpt (edam:format_2333), gro (edam:format_2033), g96 (edam:format_2033), pdb (edam:format_1476), tng (edam:format_3876).
        input_top_path (str): Path to the GROMACS input topology file. File type: input. `Sample file <https://github.com/bioexcel/biobb_analysis/raw/master/biobb_analysis/test/data/gromacs/topology.tpr>`_. Accepted formats: tpr (edam:format_2333), gro (edam:format_2033), g96 (edam:format_2033), pdb (edam:format_1476), brk (edam:format_2033), ent (edam:format_1476).
        input_index_path (str) (Optional): Path to the GROMACS index file. File type: input. `Sample file <https://github.com/bioexcel/biobb_analysis/raw/master/biobb_analysis/test/data/gromacs/index.ndx>`_. Accepted formats: ndx (edam:format_2033).
        output_str_path (str): Path to the output file. File type: output. `Sample file <https://github.com/bioexcel/biobb_analysis/raw/master/biobb_analysis/test/reference/gromacs/ref_trjconv.str.pdb>`_. Accepted formats: pdb (edam:format_1476), xtc (edam:format_3875), trr (edam:format_3910), cpt (edam:format_2333), gro (edam:format_2033), g96 (edam:format_2033), tng (edam:format_3876).
        properties (dic - Python dictionary object containing the tool parameters, not input/output files):
            * **selection** (*str*) - ("System") Group where the trjconv will be performed. If **input_index_path** provided, check the file for the accepted values. Values: System (all atoms in the system), Protein (all protein atoms), Protein-H (protein atoms excluding hydrogens), C-alpha (C-alpha atoms), Backbone (protein backbone atoms: N; C-alpha and C), MainChain (protein main chain atoms: N; C-alpha; C and O; including oxygens in C-terminus), MainChain+Cb (protein main chain atoms including C-beta), MainChain+H (protein main chain atoms including backbone amide hydrogens and hydrogens on the N-terminus), SideChain (protein side chain atoms: that is all atoms except N; C-alpha; C; O; backbone amide hydrogens and oxygens in C-terminus and hydrogens on the N-terminus), SideChain-H (protein side chain atoms excluding all hydrogens), Prot-Masses (protein atoms excluding dummy masses), non-Protein (all non-protein atoms), Water (water molecules), SOL (water molecules), non-Water (anything not covered by the Water group), Ion (any name matching an Ion entry in residuetypes.dat), NA (all NA atoms), CL (all CL atoms), Water_and_ions (combination of the Water and Ions groups), DNA (all DNA atoms), RNA (all RNA atoms), Protein_DNA (all Protein-DNA complex atoms), Protein_RNA (all Protein-RNA complex atoms), Protein_DNA_RNA (all Protein-DNA-RNA complex atoms), DNA_RNA (all DNA-RNA complex atoms).
            * **pbc** (*str*) - ("mol") PBC treatment (see help text for full description). Values: none (No PBC treatment), mol (Puts the center of mass of molecules in the box), res (Puts the center of mass of residues in the box), atom (Puts all the atoms in the box), nojump (Checks if atoms jump across the box and then puts them back), cluster (Clusters all the atoms in the selected index such that they are all closest to the center of mass of the cluster which is iteratively updated), whole (Only makes broken molecules whole).
            * **center** (*bool*) - (True) Center atoms in box.
            * **ur** (*str*) - ("compact") Unit-cell representation. Values: rect (It's the ordinary brick shape), tric (It's the triclinic unit cell), compact (Puts all atoms at the closest distance from the center of the box).
            * **fit** (*str*) - ("none") Fit molecule to ref structure in the structure file. Values: none, rot+trans, rotxy+transxy, translation, transxy, progressive.
            * **binary_path** (*str*) - ("gmx") Path to the GROMACS executable binary.
            * **remove_tmp** (*bool*) - (True) [WF property] Remove temporal files.
            * **restart** (*bool*) - (False) [WF property] Do not execute if output files exist.
            * **sandbox_path** (*str*) - ("./") [WF property] Parent path to the sandbox directory.
            * **container_path** (*str*) - (None) Container path definition.
            * **container_image** (*str*) - ('gromacs/gromacs:2022.2') Container image definition.
            * **container_volume_path** (*str*) - ('/tmp') Container volume path definition.
            * **container_working_dir** (*str*) - (None) Container working directory definition.
            * **container_user_id** (*str*) - (None) Container user_id definition.
            * **container_shell_path** (*str*) - ('/bin/bash') Path to default shell inside the container.

    Examples:
        This is a use example of how to use the building block from Python::

            from biobb_analysis.gromacs.gmx_trjconv_str import gmx_trjconv_str
            prop = {
                'selection': 'System'
            }
            gmx_trjconv_str(input_structure_path='/path/to/myStructure.trr',
                            input_top_path='/path/to/myTopology.tpr',
                            output_str_path='/path/to/newStructure.pdb',
                            input_index_path='/path/to/myIndex.ndx',
                            properties=prop)

    Info:
        * wrapped_software:
            * name: GROMACS trjconv
            * version: >=2019.1
            * license: LGPL 2.1
        * ontology:
            * name: EDAM
            * schema: http://edamontology.org/EDAM.owl

    """

    def __init__(self, input_structure_path, input_top_path, output_str_path,
                 input_index_path=None, properties=None, **kwargs) -> None:
        properties = properties or {}

        # Call parent class constructor
        super().__init__(properties)
        self.locals_var_dict = locals().copy()

        # Input/Output files
        self.io_dict = {
            "in": {"input_structure_path": input_structure_path, "input_top_path": input_top_path, "input_index_path": input_index_path},
            "out": {"output_str_path": output_str_path}
        }

        # Properties specific for BB
        self.selection = properties.get('selection', "System")
        self.pbc = properties.get('pbc', None)
        self.center = properties.get('dista', None)
        self.ur = properties.get('ur', None)
        self.fit = properties.get('fit', None)
        self.properties = properties

        # Properties common in all GROMACS BB
        self.binary_path = get_binary_path(properties, 'binary_path')

        # Check the properties
        self.check_properties(properties)
        self.check_arguments()

    def check_data_params(self, out_log, err_log):
        """ Checks all the input/output paths and parameters """
        self.io_dict["in"]["input_structure_path"] = check_traj_path(self.io_dict["in"]["input_structure_path"], out_log, self.__class__.__name__)
        self.io_dict["in"]["input_top_path"] = check_input_path(self.io_dict["in"]["input_top_path"], out_log, self.__class__.__name__)
        self.io_dict["in"]["input_index_path"] = check_index_path(self.io_dict["in"]["input_index_path"], out_log, self.__class__.__name__)
        self.io_dict["out"]["output_str_path"] = check_out_traj_path(self.io_dict["out"]["output_str_path"], out_log, self.__class__.__name__)
        if not self.io_dict["in"]["input_index_path"]:
            self.selection = get_selection(self.properties, out_log, self.__class__.__name__)
        else:
            self.selection = get_selection_index_file(self.properties, self.io_dict["in"]["input_index_path"], 'selection', out_log, self.__class__.__name__)

    @launchlogger
    def launch(self) -> int:
        """Execute the :class:`GMXTrjConvStr <gromacs.gmx_trjconv_str.GMXTrjConvStr>` gromacs.gmx_trjconv_str.GMXTrjConvStr object."""

        # check input/output paths and parameters
        self.check_data_params(self.out_log, self.err_log)

        # Setup Biobb
        if self.check_restart():
            return 0

        # standard input
        self.io_dict['in']['stdin_file_path'] = fu.create_stdin_file(f'{self.selection}')
        self.stage_files()

        self.cmd = [self.binary_path, 'trjconv',
                    '-f', self.stage_io_dict["in"]["input_structure_path"],
                    '-s', self.stage_io_dict["in"]["input_top_path"],
                    '-o', self.stage_io_dict["out"]["output_str_path"]]

        if self.pbc:
            self.cmd.append('-pbc')
            self.cmd.append(self.pbc)
        self.cmd.append('-center' if self.center else '-nocenter')
        if self.ur:
            self.cmd.append('-ur')
            self.cmd.append(self.ur)
        if self.fit:
            self.cmd.append('-fit')
            self.cmd.append(self.fit)

        if self.stage_io_dict["in"].get("input_index_path"):
            self.cmd.extend(['-n', self.stage_io_dict["in"]["input_index_path"]])

        # Add stdin input file
        self.cmd.append('<')
        self.cmd.append(self.stage_io_dict["in"]["stdin_file_path"])

        # Run Biobb block
        self.run_biobb()

        # Copy files to host
        self.copy_to_host()

        self.tmp_files.extend([
            self.stage_io_dict.get("unique_dir", ""),
            self.io_dict['in'].get("stdin_file_path", "")
        ])
        self.remove_tmp_files()

        self.check_arguments(output_files_created=True, raise_exception=False)
        return self.return_code


def gmx_trjconv_str(input_structure_path: str, input_top_path: str, output_str_path: str, input_index_path: Optional[str] = None, properties: Optional[dict] = None, **kwargs) -> int:
    """Execute the :class:`GMXTrjConvStr <gromacs.gmx_trjconv_str.GMXTrjConvStr>` class and
    execute the :meth:`launch() <gromacs.gmx_trjconv_str.GMXTrjConvStr.launch>` method."""

    return GMXTrjConvStr(input_structure_path=input_structure_path,
                         input_top_path=input_top_path,
                         output_str_path=output_str_path,
                         input_index_path=input_index_path,
                         properties=properties, **kwargs).launch()


def main():
    """Command line execution of this building block. Please check the command line documentation."""
    parser = argparse.ArgumentParser(description="Converts between GROMACS compatible structure file formats and/or extracts a selection of atoms.", formatter_class=lambda prog: argparse.RawTextHelpFormatter(prog, width=99999))
    parser.add_argument('--config', required=False, help='Configuration file')

    # Specific args of each building block
    required_args = parser.add_argument_group('required arguments')
    required_args.add_argument('--input_structure_path', required=True, help='Path to the input structure file. Accepted formats: xtc, trr, cpt, gro, g96, pdb, tng.')
    required_args.add_argument('--input_top_path', required=True, help='Path to the GROMACS input topology file. Accepted formats: tpr, gro, g96, pdb, brk, ent.')
    parser.add_argument('--input_index_path', required=False, help="Path to the GROMACS index file. Accepted formats: ndx.")
    required_args.add_argument('--output_str_path', required=True, help='Path to the output file. Accepted formats: xtc, trr, gro, g96, pdb, tng.')

    args = parser.parse_args()
    args.config = args.config or "{}"
    properties = settings.ConfReader(config=args.config).get_prop_dic()

    # Specific call of each building block
    gmx_trjconv_str(input_structure_path=args.input_structure_path,
                    input_top_path=args.input_top_path,
                    output_str_path=args.output_str_path,
                    input_index_path=args.input_index_path,
                    properties=properties)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

"""Module containing the GMX Cluster class and the command line interface."""
import argparse
from typing import Optional
from biobb_common.generic.biobb_object import BiobbObject
from biobb_common.configuration import settings
from biobb_common.tools import file_utils as fu
from biobb_common.tools.file_utils import launchlogger
from biobb_analysis.gromacs.common import check_input_path, check_traj_path, check_index_path, check_out_pdb_path, get_binary_path, get_image_selection, get_selection_index_file, get_dista, get_method, get_cutoff


class GMXCluster(BiobbObject):
    """
    | biobb_analysis GMXCluster
    | Wrapper of the GROMACS cluster module for clustering structures from a given GROMACS compatible trajectory.
    | `GROMACS cluster <http://manual.gromacs.org/current/onlinehelp/gmx-cluster.html>`_ can cluster structures using several different methods. Distances between structures can be determined from a trajectory. RMS deviation after fitting or RMS deviation of atom-pair distances can be used to define the distance between structures.

    Args:
        input_structure_path (str): Path to the input structure file. File type: input. `Sample file <https://github.com/bioexcel/biobb_analysis/raw/master/biobb_analysis/test/data/gromacs/topology.tpr>`_. Accepted formats: tpr (edam:format_2333), gro (edam:format_2033), g96 (edam:format_2033), pdb (edam:format_1476), brk (edam:format_2033), ent (edam:format_1476).
        input_traj_path (str): Path to the GROMACS trajectory file. File type: input. `Sample file <https://github.com/bioexcel/biobb_analysis/raw/master/biobb_analysis/test/data/gromacs/trajectory.trr>`_. Accepted formats: xtc (edam:format_3875), trr (edam:format_3910), cpt (edam:format_2333), gro (edam:format_2033), g96 (edam:format_2033), pdb (edam:format_1476), tng (edam:format_3876).
        input_index_path (str) (Optional): Path to the GROMACS index file. File type: input. `Sample file <https://github.com/bioexcel/biobb_analysis/raw/master/biobb_analysis/test/data/gromacs/index.ndx>`_. Accepted formats: ndx (edam:format_2033).
        output_pdb_path (str): Path to the output cluster file. File type: output. `Sample file <https://github.com/bioexcel/biobb_analysis/raw/master/biobb_analysis/test/reference/gromacs/ref_cluster.pdb>`_. Accepted formats: xtc (edam:format_3875), trr (edam:format_3910), cpt (edam:format_2333), gro (edam:format_2033), g96 (edam:format_2033), pdb (edam:format_1476), tng (edam:format_3876).
        output_cluster_log_path (str) (Optional): Path to the output log file. File type: output. `Sample file <https://github.com/bioexcel/biobb_analysis/raw/master/biobb_analysis/test/reference/gmx_cluster_cluster.log>`_. Accepted formats: log (edam:format_2330).
        output_rmsd_cluster_xpm_path (str) (Optional):Path to the output X PixMap compatible matrix file. File type: output. `Sample file <https://github.com/bioexcel/biobb_analysis/raw/master/biobb_analysis/test/reference/gmx_cluster_rmsd-clust.xpm>`_. Accepted formats: xpm (edam:format_3599).
        output_rmsd_dist_xvg_path (str) (Optional): Path to xvgr/xmgr file. File type: output. `Sample file <https://github.com/bioexcel/biobb_analysis/raw/master/biobb_analysis/test/reference/gmx_cluster_rmsd-dist.xvg>`_. Accepted formats: xvg (edam:format_2330).
        properties (dic - Python dictionary object containing the tool parameters, not input/output files):
            * **fit_selection** (*str*) - ("System") Group where the fitting will be performed. If **input_index_path** provided, check the file for the accepted values. Values: System (all atoms in the system), Protein (all protein atoms), Protein-H (protein atoms excluding hydrogens), C-alpha (C-alpha atoms), Backbone (protein backbone atoms: N; C-alpha and C), MainChain (protein main chain atoms: N; C-alpha; C and O; including oxygens in C-terminus), MainChain+Cb (protein main chain atoms including C-beta), MainChain+H (protein main chain atoms including backbone amide hydrogens and hydrogens on the N-terminus), SideChain (protein side chain atoms: that is all atoms except N; C-alpha; C; O; backbone amide hydrogens and oxygens in C-terminus and hydrogens on the N-terminus), SideChain-H (protein side chain atoms excluding all hydrogens), Prot-Masses (protein atoms excluding dummy masses), non-Protein (all non-protein atoms), Water (water molecules), SOL (water molecules), non-Water (anything not covered by the Water group), Ion (any name matching an Ion entry in residuetypes.dat), NA (all NA atoms), CL (all CL atoms), Water_and_ions (combination of the Water and Ions groups).
            * **output_selection** (*str*) - ("System") Group that is going to be written in the output trajectory. If **input_index_path** provided, check the file for the accepted values. Values: System (all atoms in the system), Protein (all protein atoms), Protein-H (protein atoms excluding hydrogens), C-alpha (C-alpha atoms), Backbone (protein backbone atoms: N; C-alpha and C), MainChain (protein main chain atoms: N; C-alpha; C and O; including oxygens in C-terminus), MainChain+Cb (protein main chain atoms including C-beta), MainChain+H (protein main chain atoms including backbone amide hydrogens and hydrogens on the N-terminus), SideChain (protein side chain atoms: that is all atoms except N; C-alpha; C; O; backbone amide hydrogens and oxygens in C-terminus and hydrogens on the N-terminus), SideChain-H (protein side chain atoms excluding all hydrogens), Prot-Masses (protein atoms excluding dummy masses), non-Protein (all non-protein atoms), Water (water molecules), SOL (water molecules), non-Water (anything not covered by the Water group), Ion (any name matching an Ion entry in residuetypes.dat), NA (all NA atoms), CL (all CL atoms), Water_and_ions (combination of the Water and Ions groups).
            * **dista** (*bool*) - (False) Use RMSD of distances instead of RMS deviation.
            * **nofit** (*bool*) - (False) Do not use least squares fitting before RMSD calculation.
            * **method** (*str*) - ("linkage") Method for cluster determination. Values: linkage (Add a structure to a cluster when its distance to any element of the cluster is less than cutoff), jarvis-patrick (Add a structure to a cluster when this structure and a structure in the cluster have each other as neighbors and they have a least P neighbors in common), monte-carlo (Reorder the RMSD matrix using Monte Carlo such that the order of the frames is using the smallest possible increments), diagonalization (Diagonalize the RMSD matrix), gromos (Count number of neighbors using cut-off and take structure with largest number of neighbors with all its neighbors as cluster and eliminate it from the pool of clusters).
            * **cutoff** (*float*) - (0.1) [0~10|0.1] RMSD cut-off (nm) for two structures to be neighbor.
            * **binary_path** (*str*) - ("gmx") Path to the GROMACS executable binary.
            * **remove_tmp** (*bool*) - (True) [WF property] Remove temporal files.
            * **restart** (*bool*) - (False) [WF property] Do not execute if output files exist.
            * **sandbox_path** (*str*) - ("./") [WF property] Parent path to the sandbox directory.
            * **container_path** (*str*) - (None) Container path definition.
            * **container_image** (*str*) - ('gromacs/gromacs:2022.2') Container image definition.
            * **container_volume_path** (*str*) - ('/tmp') Container volume path definition.
            * **container_working_dir** (*str*) - (None) Container working directory definition.
            * **container_user_id** (*str*) - (None) Container user_id definition.
            * **container_shell_path** (*str*) - ('/bin/bash') Path to default shell inside the container.

    Examples:
        This is a use example of how to use the building block from Python::

            from biobb_analysis.gromacs.gmx_cluster import gmx_cluster
            prop = {
                'fit_selection': 'System',
                'output_selection': 'System',
                'method': 'linkage'
            }
            gmx_cluster(input_structure_path='/path/to/myStructure.tpr',
                        input_traj_path='/path/to/myTrajectory.trr',
                        input_index_path='/path/to/myIndex.ndx',
                        output_pdb_path='/path/to/newStructure.pdb',
                        properties=prop)

    Info:
        * wrapped_software:
            * name: GROMACS cluster
            * version: >=2019.1
            * license: LGPL 2.1
        * ontology:
            * name: EDAM
            * schema: http://edamontology.org/EDAM.owl

    """

    def __init__(self, input_structure_path, input_traj_path, output_pdb_path,
                 input_index_path=None, output_cluster_log_path=None, output_rmsd_cluster_xpm_path=None,
                 output_rmsd_dist_xvg_path=None, properties=None, **kwargs) -> None:
        properties = properties or {}

        # Call parent class constructor
        super().__init__(properties)
        self.locals_var_dict = locals().copy()

        # Input/Output files
        self.io_dict = {
            "in": {"input_structure_path": input_structure_path, "input_traj_path": input_traj_path, "input_index_path": input_index_path},
            "out": {"output_pdb_path": output_pdb_path, "output_cluster_log_path": output_cluster_log_path,
                    "output_rmsd_cluster_xpm_path": output_rmsd_cluster_xpm_path, "output_rmsd_dist_xvg_path": output_rmsd_dist_xvg_path}
        }

        # Properties specific for BB
        self.fit_selection = properties.get('fit_selection', "System")
        self.output_selection = properties.get('output_selection', "System")
        self.method = properties.get('method', "linkage")
        self.dista = properties.get('dista', False)
        self.nofit = properties.get('nofit', False)
        self.cutoff = properties.get('cutoff', 0.1)
        self.properties = properties

        # Properties common in all GROMACS BB
        self.binary_path = get_binary_path(properties, 'binary_path')

        # Check the properties
        self.check_properties(properties)
        self.check_arguments()

    def check_data_params(self, out_log, err_log):
        """ Checks all the input/output paths and parameters """
        self.io_dict["in"]["input_structure_path"] = check_input_path(self.io_dict["in"]["input_structure_path"], out_log, self.__class__.__name__)
        self.io_dict["in"]["input_traj_path"] = check_traj_path(self.io_dict["in"]["input_traj_path"], out_log, self.__class__.__name__)
        self.io_dict["in"]["input_index_path"] = check_index_path(self.io_dict["in"]["input_index_path"], out_log, self.__class__.__name__)
        self.io_dict["out"]["output_pdb_path"] = check_out_pdb_path(self.io_dict["out"]["output_pdb_path"], out_log, self.__class__.__name__)
        if not self.io_dict["in"]["input_index_path"]:
            self.fit_selection = get_image_selection(self.properties, 'fit_selection', out_log, self.__class__.__name__)
            self.output_selection = get_image_selection(self.properties, 'output_selection', out_log, self.__class__.__name__)
        else:
            self.fit_selection = get_selection_index_file(self.properties, self.io_dict["in"]["input_index_path"], 'fit_selection', out_log, self.__class__.__name__)
            self.output_selection = get_selection_index_file(self.properties, self.io_dict["in"]["input_index_path"], 'output_selection', out_log, self.__class__.__name__)
        self.dista = get_dista(self.properties, out_log, self.__class__.__name__)
        self.method = get_method(self.properties, out_log, self.__class__.__name__)
        self.cutoff = get_cutoff(self.properties, out_log, self.__class__.__name__)

    @launchlogger
    def launch(self) -> int:
        """Execute the :class:`GMXCluster <gromacs.gmx_cluster.GMXCluster>` gromacs.gmx_cluster.GMXCluster object."""

        # check input/output paths and parameters
        self.check_data_params(self.out_log, self.err_log)

        # Optional output files
        if not self.io_dict['out'].get('output_rmsd_dist_xvg_path'):
            self.io_dict['out']['output_rmsd_dist_xvg_path'] = fu.create_name(prefix=self.prefix, step=self.step, name='rmsd-dist.xvg')
            self.tmp_files.append(self.io_dict['out']['output_rmsd_dist_xvg_path'])
        if not self.io_dict['out'].get('output_rmsd_cluster_xpm_path'):
            self.io_dict['out']['output_rmsd_cluster_xpm_path'] = fu.create_name(prefix=self.prefix, step=self.step, name='rmsd-clust.xpm')
            self.tmp_files.append(self.io_dict['out']['output_rmsd_cluster_xpm_path'])
        if not self.io_dict['out'].get('output_cluster_log_path'):
            self.io_dict['out']['output_cluster_log_path'] = fu.create_name(prefix=self.prefix, step=self.step, name='cluster.log')
            self.tmp_files.append(self.io_dict['out']['output_cluster_log_path'])

        # Setup Biobb
        if self.check_restart():
            return 0

        # standard input
        self.io_dict['in']['stdin_file_path'] = fu.create_stdin_file(f'{self.fit_selection} {self.output_selection}')
        self.stage_files()

        self.cmd = [self.binary_path, 'cluster',
                    '-g', self.stage_io_dict['out']['output_cluster_log_path'],
                    '-dist', self.stage_io_dict['out']['output_rmsd_dist_xvg_path'],
                    '-o', self.stage_io_dict['out']['output_rmsd_cluster_xpm_path'],
                    '-s', self.stage_io_dict["in"]["input_structure_path"],
                    '-f', self.stage_io_dict["in"]["input_traj_path"],
                    '-cl', self.stage_io_dict["out"]["output_pdb_path"],
                    '-cutoff', str(self.cutoff),
                    '-method', self.method]

        if self.stage_io_dict["in"].get("input_index_path"):
            self.cmd.extend(['-n', self.stage_io_dict["in"]["input_index_path"]])

        if self.dista:
            self.cmd.append('-dista')

        if self.nofit:
            self.cmd.append('-nofit')

        # Add stdin input file
        self.cmd.append('<')
        self.cmd.append(self.stage_io_dict["in"]["stdin_file_path"])

        # Run Biobb block
        self.run_biobb()

        # Copy files to host
        self.copy_to_host()

        self.tmp_files.extend([
            self.stage_io_dict.get("unique_dir", ""),
            self.io_dict['in'].get("stdin_file_path", ""),
        ])
        self.remove_tmp_files()

        self.check_arguments(output_files_created=True, raise_exception=False)

        return self.return_code


def gmx_cluster(input_structure_path: str, input_traj_path: str, output_pdb_path: str, input_index_path: Optional[str] = None, output_cluster_log_path: Optional[str] = None,
                output_rmsd_cluster_xpm_path: Optional[str] = None, output_rmsd_dist_xvg_path: Optional[str] = None, properties: Optional[dict] = None, **kwargs) -> int:
    """Execute the :class:`GMXCluster <gromacs.gmx_cluster.GMXCluster>` class and
    execute the :meth:`launch() <gromacs.gmx_cluster.GMXCluster.launch>` method."""

    return GMXCluster(input_structure_path=input_structure_path,
                      input_traj_path=input_traj_path,
                      output_pdb_path=output_pdb_path,
                      input_index_path=input_index_path,
                      output_cluster_log_path=output_cluster_log_path,
                      output_rmsd_cluster_xpm_path=output_rmsd_cluster_xpm_path,
                      output_rmsd_dist_xvg_path=output_rmsd_dist_xvg_path,
                      properties=properties, **kwargs).launch()


def main():
    """Command line execution of this building block. Please check the command line documentation."""
    parser = argparse.ArgumentParser(description="Creates cluster structures from a given GROMACS compatible trajectory.", formatter_class=lambda prog: argparse.RawTextHelpFormatter(prog, width=99999))
    parser.add_argument('--config', required=False, help='Configuration file')

    # Specific args of each building block
    required_args = parser.add_argument_group('required arguments')
    required_args.add_argument('--input_structure_path', required=True, help='Path to the input structure file. Accepted formats: tpr, gro, g96, pdb, brk, ent.')
    required_args.add_argument('--input_traj_path', required=True, help='Path to the GROMACS trajectory file. Accepted formats: xtc, trr, cpt, gro, g96, pdb, tng.')
    parser.add_argument('--input_index_path', required=False, help="Path to the GROMACS index file. Accepted formats: ndx.")
    required_args.add_argument('--output_pdb_path', required=True, help='Path to the output cluster file. Accepted formats: xtc, trr, cpt, gro, g96, pdb, tng.')
    parser.add_argument('--output_cluster_log_path', required=False, help='Path to the output log file. Accepted formats: log.')
    parser.add_argument('--output_rmsd_cluster_xpm_path', required=False, help='Path to the output xpm file. Accepted formats: xpm.')
    parser.add_argument('--output_rmsd_dist_xvg_path', required=False, help='Path to the output xvg file. Accepted formats: xvg.')

    args = parser.parse_args()
    args.config = args.config or "{}"
    properties = settings.ConfReader(config=args.config).get_prop_dic()

    # Specific call of each building block
    gmx_cluster(input_structure_path=args.input_structure_path,
                input_traj_path=args.input_traj_path,
                output_pdb_path=args.output_pdb_path,
                input_index_path=args.input_index_path,
                output_cluster_log_path=args.output_cluster_log_path,
                output_rmsd_cluster_xpm_path=args.output_rmsd_cluster_xpm_path,
                output_rmsd_dist_xvg_path=args.output_rmsd_dist_xvg_path,
                properties=properties)


if __name__ == '__main__':
    main()
